// Copyright (c) 2019 The Bitcoin SV developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#include "config.h"
#include "key.h"
#include "pubkey.h"
#include "test/test_bitcoin.h"
#include "txmempool.h"
#include "txn_validator.h"

#include <algorithm>
#include <boost/test/unit_test.hpp>

namespace {
    std::vector<TxSource> vTxSources {
        TxSource::wallet,
        TxSource::rpc,
        TxSource::file,
        TxSource::p2p,
        TxSource::reorg,
        TxSource::unknown,
        TxSource::finalised
    };
    // An exception thrown by GetValueOut .
    bool GetValueOutException(const std::runtime_error &e) {
        static std::string expectedException("GetValueOut: value out of range");
        return expectedException == e.what();
    }
    // Support for P2P node.
    CService ip(uint32_t i) {
        struct in_addr s;
        s.s_addr = i;
        return CService(CNetAddr(s), Params().GetDefaultPort());
    }
    // Create scriptPubKey from a given key
    CScript GetScriptPubKey(CKey& key) {
        CScript scriptPubKey = CScript() << ToByteVector(key.GetPubKey())
                                         << OP_CHECKSIG;
        return scriptPubKey;
    }
    // Create a double spend txn
    CMutableTransaction CreateDoubleSpendTxn(const CTransaction& fundTxn,
                                       CKey& key,
                                       CScript& scriptPubKey) {
        static uint32_t dummyLockTime = 0;
        CMutableTransaction spend_txn;
        spend_txn.nVersion = 1;
        spend_txn.nLockTime = ++dummyLockTime;
        spend_txn.vin.resize(1);
        spend_txn.vin[0].prevout = COutPoint(fundTxn.GetId(), 0);
        spend_txn.vout.resize(1);
        spend_txn.vout[0].nValue = 11 * CENT;
        spend_txn.vout[0].scriptPubKey = scriptPubKey;
        // Sign:
        std::vector<uint8_t> vchSig {};
        uint256 hash = SignatureHash(scriptPubKey, CTransaction(spend_txn), 0,
                                     SigHashType().withForkId(),
                                     fundTxn.vout[0].nValue);
        BOOST_CHECK(key.Sign(hash, vchSig));
        vchSig.push_back(uint8_t(SIGHASH_ALL | SIGHASH_FORKID));
        spend_txn.vin[0].scriptSig << vchSig;
        return spend_txn;
    }
    // Make N unique large (but rubbish) transactions
    std::vector<CMutableTransaction> MakeNLargeTxns(size_t nNumTxns,
                                                    CTransaction& fundTxn,
                                                    CScript& scriptPubKey) {
        std::vector<CMutableTransaction> res {};
        for (size_t i=0; i<nNumTxns; i++) {
            CMutableTransaction txn;
            txn.nVersion = 1;
            txn.vin.resize(1);
            txn.vin[0].prevout = COutPoint(fundTxn.GetId(), i);
            txn.vout.resize(1000);
            for(size_t j=0; j<1000; ++j) {
                txn.vout[j].nValue = 11 * CENT;
                txn.vout[j].scriptPubKey = scriptPubKey;
            }
            res.emplace_back(txn);
        }
        return res;
    }
    // Create N double spend txns from the given fund txn
    std::vector<CMutableTransaction> CreateNDoubleSpendTxns(size_t nSpendTxns,
                                                      CTransaction& fundTxn,
                                                      CKey& key,
                                                      CScript& scriptPubKey) {
        // Create txns spending the same coinbase txn.
        std::vector<CMutableTransaction> spends {};
        for (size_t i=0; i<nSpendTxns; i++) {
            spends.emplace_back(CreateDoubleSpendTxn(fundTxn, key, scriptPubKey));
        };
        return spends;
    }
    // Create txn input data for a given txn and source
    TxInputDataSPtr TxInputData(TxSource source,
                                CMutableTransaction& spend,
                                std::shared_ptr<CNode> pNode = nullptr) {
        // Return txn's input data
        return std::make_shared<CTxInputData>(
                   g_connman->GetTxIdTracker(), // a pointer to the TxIdTracker
                   MakeTransactionRef(spend),// a pointer to the tx
                   source,   // tx source
                   TxValidationPriority::normal, // tx validation priority
                   GetTime(),// nAcceptTime
                   false,    // mfLimitFree
                   Amount(0), // nAbsurdFee
                   pNode);   // pNode
    }
    // Create a vector with input data for a given txn and source
    std::vector<TxInputDataSPtr> TxInputDataVec(TxSource source,
                                                const std::vector<CMutableTransaction>& spends,
                                                std::shared_ptr<CNode> pNode = nullptr) {
        std::vector<TxInputDataSPtr> vTxInputData {};
        // Get a pointer to the TxIdTracker.
        const TxIdTrackerSPtr& pTxIdTracker = g_connman->GetTxIdTracker();
        for (auto& elem : spends) {
            vTxInputData.
                emplace_back(
                    std::make_shared<CTxInputData>(
                        pTxIdTracker, // a pointer to the TxIdTracker
                        MakeTransactionRef(elem),  // a pointer to the tx
                        source,   // tx source
                        TxValidationPriority::normal, // tx validation priority
                        GetTime(),// nAcceptTime
                        false,    // mfLimitFree
                        Amount(0), // nAbsurdFee
                        pNode));   // pNode
        }
        return vTxInputData;
    }
    // Validate txn using asynchronous validation interface
    void ProcessTxnsAsynchApi(std::vector<CMutableTransaction>& spends,
                              TxSource source,
                              std::shared_ptr<CNode> pNode = nullptr) {
        // Create txn validator
        std::shared_ptr<CTxnValidator> txnValidator {
            std::make_shared<CTxnValidator>(
                    GlobalConfig::GetConfig(),
                    mempool,
                    std::make_shared<CTxnDoubleSpendDetector>(),
                    g_connman->GetTxIdTracker())
        };
        // Clear mempool before validation
        mempool.Clear();
        // Schedule txns for processing.
        txnValidator->newTransaction(TxInputDataVec(source, spends, pNode));
        // Wait for the Validator to process all queued txns.
        txnValidator->waitForEmptyQueue();
    }
    // Validate a single txn using synchronous validation interface
    CValidationState ProcessTxnSynchApi(CMutableTransaction& spend,
                                        TxSource source,
                                        std::shared_ptr<CNode> pNode = nullptr) {
        // Create txn validator
        std::shared_ptr<CTxnValidator> txnValidator {
            std::make_shared<CTxnValidator>(
                    GlobalConfig::GetConfig(),
                    mempool,
                    std::make_shared<CTxnDoubleSpendDetector>(),
                    g_connman->GetTxIdTracker())
        };
        // Clear mempool before validation
        mempool.Clear();
        // Mempool Journal ChangeSet
        mining::CJournalChangeSetPtr changeSet {nullptr};
        return txnValidator->processValidation(TxInputData(source, spend, pNode), changeSet);
    }
    // Validate txn using synchronous validation interface
    void ProcessTxnsSynchApi(std::vector<CMutableTransaction>& spends,
                             TxSource source,
                             std::shared_ptr<CNode> pNode = nullptr) {
        // Create txn validator
        std::shared_ptr<CTxnValidator> txnValidator {
            std::make_shared<CTxnValidator>(
                    GlobalConfig::GetConfig(),
                    mempool,
                    std::make_shared<CTxnDoubleSpendDetector>(),
                    g_connman->GetTxIdTracker())
        };
        // Clear mempool before validation
        mempool.Clear();
        // Mempool Journal ChangeSet
        mining::CJournalChangeSetPtr changeSet {nullptr};
        // Validate the first txn
        CValidationState result {
            txnValidator->processValidation(TxInputData(source, spends[0], pNode), changeSet)
        };
        BOOST_CHECK(result.IsValid());
        // Validate the second txn
        // spends1 should be rejected if spend0 is in the mempool
        result = txnValidator->processValidation(TxInputData(source, spends[1], pNode), changeSet);
        BOOST_CHECK(!result.IsValid());
    }
    // Validate txns using synchronous batch validation interface
    CTxnValidator::RejectedTxns ProcessTxnsSynchBatchApi(std::vector<CMutableTransaction>& spends,
                                  TxSource source,
                                  std::shared_ptr<CNode> pNode = nullptr) {
        // Create txn validator
        std::shared_ptr<CTxnValidator> txnValidator {
            std::make_shared<CTxnValidator>(
                    GlobalConfig::GetConfig(),
                    mempool,
                    std::make_shared<CTxnDoubleSpendDetector>(),
                    g_connman->GetTxIdTracker())
        };
        // Clear mempool before validation
        mempool.Clear();
        // Mempool Journal ChangeSet
        mining::CJournalChangeSetPtr changeSet {nullptr};
        // Validate the first txn
        return txnValidator->processValidation(TxInputDataVec(source, spends, pNode), changeSet);
    }
    struct TestChain100Setup2 : TestChain100Setup {
        CScript scriptPubKey {
            GetScriptPubKey(coinbaseKey)
        };
        // twoDoubleSpend2Txns contains two txns spending the same coinbase txn
        std::vector<CMutableTransaction> doubleSpend2Txns {
            CreateDoubleSpendTxn(coinbaseTxns[0], coinbaseKey, scriptPubKey),
            CreateDoubleSpendTxn(coinbaseTxns[0], coinbaseKey, scriptPubKey)
        };
        // doubleSpend10Txns contains 10 double spend txns spending the same coinbase txn
        std::vector<CMutableTransaction> doubleSpend10Txns {
            CreateNDoubleSpendTxns(10, coinbaseTxns[0], coinbaseKey, scriptPubKey)
        };    
    };
}

BOOST_FIXTURE_TEST_SUITE(test_txnvalidator, TestChain100Setup2)

BOOST_AUTO_TEST_CASE(txn_validator_creation) {
    // Create txn validator
    std::shared_ptr<CTxnValidator> txnValidator {
        std::make_shared<CTxnValidator>(
                GlobalConfig::GetConfig(),
                mempool,
                std::make_shared<CTxnDoubleSpendDetector>(),
                g_connman->GetTxIdTracker())
    };
    // Check if the Validator was created
    BOOST_REQUIRE(txnValidator);
    // Check if orphan txns buffer was created
    BOOST_REQUIRE(txnValidator->getOrphanTxnsPtr());
    // Check if txn recent rejects buffer was created
    BOOST_REQUIRE(txnValidator->getTxnRecentRejectsPtr());
}

BOOST_AUTO_TEST_CASE(txn_validator_set_get_frequency) {
    // Create txn validator
    std::shared_ptr<CTxnValidator> txnValidator {
        std::make_shared<CTxnValidator>(
                GlobalConfig::GetConfig(),
                mempool,
                std::make_shared<CTxnDoubleSpendDetector>(),
                g_connman->GetTxIdTracker())
    };
    auto defaultfreq = std::chrono::milliseconds(CTxnValidator::DEFAULT_ASYNCH_RUN_FREQUENCY_MILLIS);
    BOOST_CHECK(defaultfreq == txnValidator->getRunFrequency());
    txnValidator->setRunFrequency(++defaultfreq);
    BOOST_CHECK(defaultfreq == txnValidator->getRunFrequency());
}

BOOST_AUTO_TEST_CASE(txn_validator_istxnknown) {
    // Create txn validator
    std::shared_ptr<CTxnValidator> txnValidator {
        std::make_shared<CTxnValidator>(
                GlobalConfig::GetConfig(),
                mempool,
                std::make_shared<CTxnDoubleSpendDetector>(),
                g_connman->GetTxIdTracker())
    };
    // Schedule txns for processing.
    txnValidator->newTransaction(TxInputDataVec(TxSource::p2p, doubleSpend10Txns));
    BOOST_CHECK(txnValidator->isTxnKnown(doubleSpend10Txns[0].GetId()));
    // Wait for the Validator to process all queued txns.
    txnValidator->waitForEmptyQueue();
    BOOST_CHECK(!txnValidator->isTxnKnown(doubleSpend10Txns[0].GetId()));
}

BOOST_AUTO_TEST_CASE(double_spend_detector)
{
    CTxnDoubleSpendDetector detector;
    std::vector<CMutableTransaction> txns{ MakeNLargeTxns(5, coinbaseTxns[0], scriptPubKey) };
    std::size_t parentOfDoubleSpendIdx = 1;
    // replace the transaction that spends txns[parentOfDoubleSpendIdx] so that we can crate a double spend later
    txns[2] = CreateDoubleSpendTxn(CTransaction{txns[parentOfDoubleSpendIdx]}, coinbaseKey, scriptPubKey);
    auto txnsData = TxInputDataVec(TxSource::p2p, txns);

    for(const auto& data : txnsData)
    {
        CValidationState state;
        BOOST_CHECK(detector.insertTxnInputs(data, mempool, state, true) == true);

        BOOST_CHECK(state.IsDoubleSpendDetected() == false);
        BOOST_CHECK(state.IsMempoolConflictDetected() == false);
        BOOST_CHECK(state.GetCollidedWithTx().empty());
    }

    std::size_t doubleSpendIdx = 2;
    auto doubleSpendTx =
        std::make_shared<CTransaction>(
            CreateDoubleSpendTxn(CTransaction{txns[parentOfDoubleSpendIdx]}, coinbaseKey, scriptPubKey) );
    auto doubleSpendData =
        std::make_shared<CTxInputData>(
            g_connman->GetTxIdTracker(),
            doubleSpendTx,
            TxSource::p2p,
            TxValidationPriority::normal,
            GetTime(),// nAcceptTime
            false,    // mfLimitFree
            Amount(0), // nAbsurdFee
            std::weak_ptr<CNode>{});
    auto& primaryTx = *txnsData[doubleSpendIdx]->GetTxnPtr();

    // Double spend should be detected
    {
        CValidationState state;
        BOOST_CHECK(
            detector.insertTxnInputs(txnsData[doubleSpendIdx], mempool, state, true)
            == false);
        BOOST_CHECK(state.IsDoubleSpendDetected() == true);
        BOOST_CHECK(state.IsMempoolConflictDetected() == false);
        BOOST_CHECK_EQUAL(state.GetCollidedWithTx().size(), 1);
        BOOST_CHECK_EQUAL((*state.GetCollidedWithTx().begin())->GetId().ToString(), primaryTx.GetId().ToString());
    }

    // Trying to remove the tx with different address doesn't change anything
    {
        detector.removeTxnInputs( CTransaction{txns[doubleSpendIdx]} );
        CValidationState state;
        BOOST_CHECK(
            detector.insertTxnInputs(doubleSpendData, mempool, state, true)
            == false);
        BOOST_CHECK(state.IsDoubleSpendDetected() == true);
        BOOST_CHECK(state.IsMempoolConflictDetected() == false);
        BOOST_CHECK_EQUAL(state.GetCollidedWithTx().size(), 1);
        BOOST_CHECK_EQUAL((*state.GetCollidedWithTx().begin())->GetId().ToString(), primaryTx.GetId().ToString());
    }

    // Trying to remove the double spend tx doesn't change anything
    {
        detector.removeTxnInputs(*doubleSpendTx);
        CValidationState state;
        BOOST_CHECK(
            detector.insertTxnInputs(doubleSpendData, mempool, state, true)
            == false);
        BOOST_CHECK(state.IsDoubleSpendDetected() == true);
        BOOST_CHECK(state.IsMempoolConflictDetected() == false);
        BOOST_CHECK_EQUAL(state.GetCollidedWithTx().size(), 1);
        BOOST_CHECK_EQUAL((*state.GetCollidedWithTx().begin())->GetId().ToString(), primaryTx.GetId().ToString());
    }

    // Remove the first tx that caused double spend spend transaction so that we
    // can add double spend tx without error
    {
        detector.removeTxnInputs(*txnsData[doubleSpendIdx]->GetTxnPtr());
        CValidationState state;
        BOOST_CHECK(detector.insertTxnInputs(doubleSpendData, mempool, state, true) == true);

        BOOST_CHECK(state.IsDoubleSpendDetected() == false);
        BOOST_CHECK(state.IsMempoolConflictDetected() == false);
        BOOST_CHECK(state.GetCollidedWithTx().empty());
    }

    // Remove the double spend transaction, add initial tx to mempool and make
    // sure that a mempool collision is detected when adding the double spend tx
    // to the detector
    {
        detector.removeTxnInputs(*doubleSpendTx);

        Amount fee{ 3 };
        int64_t time = 0;
        double priority = 10.0;
        unsigned int height = 1;
        bool spendsCoinbase = false;
        unsigned int sigOpCost = 4;
        LockPoints lp;
        mining::CJournalChangeSetPtr nullChangeSet{nullptr};
        auto& tx = *txnsData[doubleSpendIdx]->GetTxnPtr();
        mempool.AddUnchecked(
            tx.GetId(),
            CTxMemPoolEntry{
                txnsData[doubleSpendIdx]->GetTxnPtr(),
                fee,
                time,
                priority,
                height,
                tx.GetValueOut(),
                spendsCoinbase,
                sigOpCost,
                lp},
            nullChangeSet);

        CValidationState state;
        BOOST_CHECK(
            detector.insertTxnInputs(doubleSpendData, mempool, state, true)
            == false);
        BOOST_CHECK(state.IsDoubleSpendDetected() == false);
        BOOST_CHECK(state.IsMempoolConflictDetected() == true);
        BOOST_CHECK_EQUAL(state.GetCollidedWithTx().size(), 1);
        BOOST_CHECK_EQUAL((*state.GetCollidedWithTx().begin())->GetId().ToString(), primaryTx.GetId().ToString());
    }
}

BOOST_AUTO_TEST_CASE(validation_state_collided_with_tx)
{
    std::vector<CMutableTransaction> txns{ MakeNLargeTxns(7, coinbaseTxns[0], scriptPubKey) };
    CValidationState state;

    auto all_present =
        [](
            const std::set<CTransactionRef>& transactions,
            const std::vector<CTransactionRef>& expected)
        {
            BOOST_CHECK_EQUAL(transactions.size(), expected.size());
            BOOST_CHECK(
                std::all_of(
                    expected.begin(),
                    expected.end(),
                    [&transactions](const CTransactionRef& item)
                    {
                        return transactions.find(item) != transactions.end();
                    }));
        };

    BOOST_CHECK(state.IsDoubleSpendDetected() == false);
    BOOST_CHECK(state.IsMempoolConflictDetected() == false);
    BOOST_CHECK_EQUAL(state.GetCollidedWithTx().size(), 0);

    std::vector<CTransactionRef> added;

    // add two transactions as double spends
    added.push_back(MakeTransactionRef(txns[0]));
    added.push_back(MakeTransactionRef(txns[1]));
    state.SetDoubleSpendDetected({added[0], added[1]});
    BOOST_CHECK(state.IsDoubleSpendDetected() == true);
    BOOST_CHECK(state.IsMempoolConflictDetected() == false);
    all_present( state.GetCollidedWithTx(), added );

    // we can call SetDoubleSpendDetected() multiple times but duplicates won't be added
    added.push_back(MakeTransactionRef(txns[2]));
    state.SetDoubleSpendDetected({added[1], added[2]});
    BOOST_CHECK(state.IsDoubleSpendDetected() == true);
    BOOST_CHECK(state.IsMempoolConflictDetected() == false);
    all_present( state.GetCollidedWithTx(), added );

    // add two transactions as mempool conflicts
    added.push_back(MakeTransactionRef(txns[3]));
    added.push_back(MakeTransactionRef(txns[4]));
    state.SetMempoolConflictDetected({added[3], added[4]});
    BOOST_CHECK(state.IsDoubleSpendDetected() == true);
    BOOST_CHECK(state.IsMempoolConflictDetected() == true);
    all_present( state.GetCollidedWithTx(), added );

    // we can call SetMempoolConflictDetected() multiple times but duplicates won't be added
    added.push_back(MakeTransactionRef(txns[5]));
    state.SetMempoolConflictDetected({added[4], added[5]});
    BOOST_CHECK(state.IsDoubleSpendDetected() == true);
    BOOST_CHECK(state.IsMempoolConflictDetected() == true);
    all_present( state.GetCollidedWithTx(), added );

    // clear the collided with container
    state.ClearCollidedWithTx();
    BOOST_CHECK(state.IsDoubleSpendDetected() == true);
    BOOST_CHECK(state.IsMempoolConflictDetected() == true);
    BOOST_CHECK_EQUAL(state.GetCollidedWithTx().size(), 0);
}

/**
 * TxnValidator: Test synch interface.
 */
BOOST_AUTO_TEST_CASE(txnvalidator_doublespend_synch_api) {
    // Test all sources.
    for (const auto& txsource: vTxSources) {
        ProcessTxnsSynchApi(doubleSpend2Txns, txsource);
        BOOST_CHECK_EQUAL(mempool.Size(), 1);
    }
    // Test: Txns from p2p with a pointer to a dummy node.
    {
        // Create a dummy address
        CAddress dummy_addr(ip(0xa0b0c001), NODE_NONE);
        CConnman::CAsyncTaskPool asyncTaskPool{GlobalConfig::GetConfig()};
        CNodePtr pDummyNode =
            CNode::Make(
                0,
                NODE_NETWORK,
                0,
                INVALID_SOCKET,
                dummy_addr,
                0u,
                0u,
                asyncTaskPool,
                "",
                true);
        ProcessTxnsSynchApi(doubleSpend2Txns, TxSource::p2p, pDummyNode);
        BOOST_CHECK_EQUAL(mempool.Size(), 1);
    }
}

BOOST_AUTO_TEST_CASE(txnvalidator_doublespend_synch_batch_api) {
    CTxnValidator::RejectedTxns mRejectedTxns {};
    // Test all sources.
    for (const auto& txsource: vTxSources) {
        mRejectedTxns = ProcessTxnsSynchBatchApi(doubleSpend10Txns, txsource);
        BOOST_CHECK_EQUAL(mempool.Size(), 1);
        // There should be no insufficient fee txns returned.
        BOOST_REQUIRE(!mRejectedTxns.second.size());
        // Check an expected number of invalid txns returned.
        const CTxnValidator::InvalidTxnStateUMap& mInvalidTxns = mRejectedTxns.first;
        BOOST_REQUIRE(mInvalidTxns.size() == doubleSpend10Txns.size()-1);
        for (const auto& elem : mInvalidTxns) {
            BOOST_REQUIRE(!elem.second.IsValid());
            // Due to runtime-conditions it might be detected as:
            // - a mempool conflict
            // - a double spend
            BOOST_REQUIRE(elem.second.IsMempoolConflictDetected() ||
                elem.second.IsDoubleSpendDetected());
        }
    }
    // Test: Txns from p2p with a pointer to a dummy node.
    {
        // Create a dummy address
        CAddress dummy_addr(ip(0xa0b0c001), NODE_NONE);
        CConnman::CAsyncTaskPool asyncTaskPool{GlobalConfig::GetConfig()};
        CNodePtr pDummyNode =
            CNode::Make(
                0,
                NODE_NETWORK,
                0,
                INVALID_SOCKET,
                dummy_addr,
                0u,
                0u,
                asyncTaskPool,
                "",
                true);
        mRejectedTxns = ProcessTxnsSynchBatchApi(doubleSpend10Txns, TxSource::p2p, pDummyNode);
        BOOST_CHECK_EQUAL(mempool.Size(), 1);
        // There should be no insufficient fee txns returned.
        BOOST_REQUIRE(!mRejectedTxns.second.size());
        // Check an expected number of invalid txns returned.
        const CTxnValidator::InvalidTxnStateUMap& mInvalidTxns = mRejectedTxns.first;
        BOOST_REQUIRE(mInvalidTxns.size() == doubleSpend10Txns.size()-1);
        for (const auto& elem : mInvalidTxns) {
            BOOST_REQUIRE(!elem.second.IsValid());
            // Due to runtime-conditions it might be detected as:
            // - a mempool conflict
            // - a double spend
            BOOST_REQUIRE(elem.second.IsMempoolConflictDetected() ||
                elem.second.IsDoubleSpendDetected());
        }
    }
}

/**
 * TxnValidator: Test asynch interface.
 */
BOOST_AUTO_TEST_CASE(txnvalidator_wallet_doublespend_via_asynch_api) {
    // Test: Txns from wallet.
    ProcessTxnsAsynchApi(doubleSpend10Txns, TxSource::wallet);
    BOOST_CHECK_EQUAL(mempool.Size(), 1);
}

BOOST_AUTO_TEST_CASE(txnvalidator_rpc_doublespend_via_asynch_api) {
    // Test: Txns from rpc.
    ProcessTxnsAsynchApi(doubleSpend10Txns, TxSource::rpc);
    BOOST_CHECK_EQUAL(mempool.Size(), 1);
}

BOOST_AUTO_TEST_CASE(txnvalidator_file_doublespend_via_asynch_api) {
    // Test: Txns from file.
    ProcessTxnsAsynchApi(doubleSpend10Txns, TxSource::file);
    BOOST_CHECK_EQUAL(mempool.Size(), 1);
}

BOOST_AUTO_TEST_CASE(txnvalidator_p2p_doublespend_via_asynch_api) {
    // Test: Txns from p2p.
    {
        // Create a dummy address
        CAddress dummy_addr(ip(0xa0b0c001), NODE_NONE);
        CConnman::CAsyncTaskPool asyncTaskPool{GlobalConfig::GetConfig()};
        CNodePtr pDummyNode =
            CNode::Make(
                0,
                NODE_NETWORK,
                0,
                INVALID_SOCKET,
                dummy_addr,
                0u,
                0u,
                asyncTaskPool,
                "",
                true);
        ProcessTxnsAsynchApi(doubleSpend10Txns, TxSource::p2p, pDummyNode);
        BOOST_CHECK_EQUAL(mempool.Size(), 1);
    }
    // Process txn if it is valid.
    ProcessTxnsAsynchApi(doubleSpend10Txns, TxSource::p2p);
    BOOST_CHECK_EQUAL(mempool.Size(), 1);
}

BOOST_AUTO_TEST_CASE(txnvalidator_reorg_doublespend_via_asynch_api) {
    // Test: Txns from reorg.
    ProcessTxnsAsynchApi(doubleSpend10Txns, TxSource::reorg);
    BOOST_CHECK_EQUAL(mempool.Size(), 1);
}

BOOST_AUTO_TEST_CASE(txnvalidator_dummy_doublespend_via_asynch_api) {
    // Process txn if it is valid.
    ProcessTxnsAsynchApi(doubleSpend10Txns, TxSource::unknown);
    BOOST_CHECK_EQUAL(mempool.Size(), 1);
}

BOOST_AUTO_TEST_CASE(txnvalidator_limit_memory_usage)
{
    // Make sure validation thread won't run during this test
    gArgs.ForceSetArg("-txnvalidationasynchrunfreq", "10000");
    gArgs.ForceSetArg("-txnvalidationqueuesmaxmemory", "1");

    // Create a larger number of txns than will fit in a 1Mb queue
    std::vector<CMutableTransaction> txns { MakeNLargeTxns(25, coinbaseTxns[0], scriptPubKey) };
    auto txnsInputs { TxInputDataVec(TxSource::p2p, txns) };

    // Create txn validator
    std::shared_ptr<CTxnValidator> txnValidator {
        std::make_shared<CTxnValidator>(
                GlobalConfig::GetConfig(),
                mempool,
                std::make_shared<CTxnDoubleSpendDetector>(),
                g_connman->GetTxIdTracker())
    };

    // Attempt to enqueue all txns and verify that we stopped when we hit the max size limit
    txnValidator->newTransaction(txnsInputs);
    BOOST_CHECK(txnValidator->GetTransactionsInQueueCount() < txns.size());
    BOOST_CHECK(txnValidator->GetStdQueueMemUsage() <= 1*ONE_MEBIBYTE);
    BOOST_CHECK_EQUAL(txnValidator->GetNonStdQueueMemUsage(), 0);
}

BOOST_AUTO_TEST_CASE(txnvalidator_nvalueoutofrange_sync_api) {
    // spendtx_nValue_OutOfRange (a copy of doubleSpend2Txns[0]) with unsupported nValue amount.
    // Set nValue = MAX_MONEY + 1 for the txn to trigger exception when GetValueOut is called.
    auto spendtx_nValue_OutOfRange = doubleSpend2Txns[0];
    spendtx_nValue_OutOfRange.vout[0].nValue = MAX_MONEY + Amount(1);
    BOOST_CHECK_EXCEPTION(
        !MoneyRange(CTransaction(spendtx_nValue_OutOfRange).GetValueOut()),
        std::runtime_error,
        GetValueOutException);
    CValidationState result {};
    // Test all sources.
    for (const auto& txsource: vTxSources) {
        result = ProcessTxnSynchApi(spendtx_nValue_OutOfRange, txsource);
        BOOST_CHECK(!result.IsValid());
        BOOST_CHECK_EQUAL(mempool.Size(), 0);
    }
}

BOOST_AUTO_TEST_CASE(txnvalidator_nvalueoutofrange_async_api) {
    // Create txn validator
    std::shared_ptr<CTxnValidator> txnValidator {
        std::make_shared<CTxnValidator>(
                GlobalConfig::GetConfig(),
                mempool,
                std::make_shared<CTxnDoubleSpendDetector>(),
                g_connman->GetTxIdTracker())
    };
    // Case1:
    // doubleSpends10Txns_nValue_OutOfRange (a copy of doubleSpend10Txns) with unsupported nValue amount.
    {
        // Set nValue = MAX_MONEY + 1 for each txn to trigger exception when GetValueOut is called.
        auto doubleSpends10Txns_nValue_OutOfRange = doubleSpend10Txns;
        for (auto& spend: doubleSpends10Txns_nValue_OutOfRange) {
            spend.vout[0].nValue = MAX_MONEY + Amount(1);
            BOOST_CHECK_EXCEPTION(!MoneyRange(CTransaction(spend).GetValueOut()), std::runtime_error, GetValueOutException);
        }
        // Schedule txns for processing.
        txnValidator->newTransaction(TxInputDataVec(TxSource::p2p, doubleSpends10Txns_nValue_OutOfRange));
        // Wait for the Validator to process all queued txns.
        txnValidator->waitForEmptyQueue();
        // Non transaction should be accepted due to nValue (value out of range).
        BOOST_CHECK_EQUAL(mempool.Size(), 0);
    }
    // Case2:
    // Send the same txns again (with valid nValue).
    // Check if only one txn (from doubleSpend10Txns) is accepted by the mempool.
    {
        txnValidator->newTransaction(TxInputDataVec(TxSource::p2p, doubleSpend10Txns));
        txnValidator->waitForEmptyQueue();
        BOOST_CHECK_EQUAL(mempool.Size(), 1);
    }
}

BOOST_AUTO_TEST_SUITE_END()
